@@ -35,7 +35,7 @@ import time
 
 import ui.emblemcellrenderer
 
-from util.namedtuple import namedtuple
+from collections import namedtuple
 from meldapp import app
 
 gdk = gtk.gdk
@@ -46,8 +46,6 @@ gdk = gtk.gdk
 #
 ################################################################################
 
-# For compatibility with Python 2.5, we use the Python 2.4 compatible version of
-# namedtuple. The class is included in the collections module as of Python 2.6.
 class StatItem(namedtuple('StatItem', 'mode size time')):
     __slots__ = ()
 
@@ -167,8 +165,7 @@ COL_EMBLEM, COL_END = tree.COL_END, tree.COL_END + 1
 ################################################################################
 class DirDiffTreeStore(tree.DiffTreeStore):
     def __init__(self, ntree):
-        types = [str] * COL_END * ntree
-        tree.DiffTreeStore.__init__(self, ntree, types)
+        tree.DiffTreeStore.__init__(self, ntree, [str])
 
 
 class CanonicalListing(object):
@@ -214,8 +211,8 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
         tree.STATE_ERROR: "error",
         tree.STATE_EMPTY: None,
         tree.STATE_MODIFIED: "replace",
-        tree.STATE_CONFLICT: "conflict",
         tree.STATE_MISSING: "delete",
+        tree.STATE_NONEXIST: "delete",
     }
 
     state_actions = {
@@ -237,7 +234,7 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
         )
 
         toggleactions = (
-            ("IgnoreCase",   gtk.STOCK_ITALIC,  _("Case"),     None, _("Ignore case of entries"), self.on_button_ignore_case_toggled, False),
+            ("IgnoreCase",   gtk.STOCK_ITALIC,  _("Ignore filename case"),     None, _("Consider differently-cased filenames that are otherwise-identical to be the same"), self.on_button_ignore_case_toggled, False),
             ("ShowSame",     gtk.STOCK_APPLY,   _("Same"),     None, _("Show identical"), self.on_filter_state_toggled, False),
             ("ShowNew",      gtk.STOCK_ADD,     _("New"),      None, _("Show new"), self.on_filter_state_toggled, False),
             ("ShowModified", gtk.STOCK_REMOVE,  _("Modified"), None, _("Show modified"), self.on_filter_state_toggled, False),
@@ -250,19 +247,29 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
         self.actiongroup.add_actions(actions)
         self.actiongroup.add_toggle_actions(toggleactions)
         self.name_filters = []
-        self.create_name_filters()
-        app.connect("file-filters-changed", self.on_file_filters_changed)
         self.text_filters = []
+        self.create_name_filters()
         self.create_text_filters()
-        app.connect("text-filters-changed", self.on_text_filters_changed)
+        self.app_handlers = [app.connect("file-filters-changed",
+                                         self.on_file_filters_changed),
+                             app.connect("text-filters-changed",
+                                         self.on_text_filters_changed)]
+
         for button in ("DirCompare", "DirCopyLeft", "DirCopyRight",
-                       "DirDelete", "Hide", "IgnoreCase", "ShowSame",
+                       "DirDelete", "ShowSame",
                        "ShowNew", "ShowModified", "CustomFilterMenu"):
             self.actiongroup.get_action(button).props.is_important = True
         self.map_widgets_into_lists(["treeview", "fileentry", "scrolledwindow",
                                      "diffmap", "linkmap", "msgarea_mgr",
                                      "vbox"])
+
+        self.widget.connect("style-set", self.on_style_set)
+        self.widget.ensure_style()
+
         self.set_num_panes(num_panes)
+
+        self.widget.connect("style-set", self.model.on_style_set)
+
         self.focus_in_events = []
         self.focus_out_events = []
         for treeview in self.treeview:
@@ -274,6 +281,14 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
         self.on_treeview_focus_out_event(None, None)
         self.treeview_focussed = None
 
+        lastchanged_label = gtk.Label()
+        lastchanged_label.set_size_request(100, -1)
+        lastchanged_label.show()
+        permissions_label = gtk.Label()
+        permissions_label.set_size_request(100, -1)
+        permissions_label.show()
+        self.status_info_labels = [lastchanged_label, permissions_label]
+
         for i in range(3):
             self.treeview[i].get_selection().set_mode(gtk.SELECTION_MULTIPLE)
             column = gtk.TreeViewColumn()
@@ -282,7 +297,11 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
             column.pack_start(renicon, expand=0)
             column.pack_start(rentext, expand=1)
             col_index = self.model.column_index
-            column.set_attributes(rentext, markup=col_index(tree.COL_TEXT,i))
+            column.set_attributes(rentext, text=col_index(tree.COL_TEXT,i),
+                                  foreground_gdk=col_index(tree.COL_FG, i),
+                                  style=col_index(tree.COL_STYLE, i),
+                                  weight=col_index(tree.COL_WEIGHT, i),
+                                  strikethrough=col_index(tree.COL_STRIKE, i))
             column.set_attributes(renicon,
                                   icon_name=col_index(tree.COL_ICON, i),
                                   emblem_name=col_index(COL_EMBLEM, i),
@@ -299,6 +318,31 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
                 action_name = self.state_actions[s][1]
                 self.actiongroup.get_action(action_name).set_active(True)
 
+    def on_style_set(self, widget, prev_style):
+        style = widget.get_style()
+
+        lookup = lambda color_id, default: style.lookup_color(color_id) or \
+                                           gtk.gdk.color_parse(default)
+
+        self.fill_colors = {"insert"  : lookup("insert-bg", "DarkSeaGreen1"),
+                            "delete"  : lookup("delete-bg", "White"),
+                            "replace" : lookup("replace-bg", "#ddeeff"),
+                            "error"   : lookup("error-bg", "#fce94f")}
+        self.line_colors = {"insert"  : lookup("insert-outline", "#77f077"),
+                            "delete"  : lookup("delete-outline", "Grey"),
+                            "replace" : lookup("replace-outline", "#8bbff3"),
+                            "error"   : lookup("error-outline", "#edd400")}
+
+        for diffmap in self.diffmap:
+            diffmap.set_color_scheme([self.fill_colors, self.line_colors])
+        self.queue_draw()
+
+    def queue_draw(self):
+        for treeview in self.treeview:
+            treeview.queue_draw()
+        for diffmap in self.diffmap:
+            diffmap.queue_draw()
+
     def on_custom_filter_menu_toggled(self, item):
         if item.get_active():
             self.custom_popup.connect("deactivate",
@@ -437,6 +481,7 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
     def set_locations(self, locations):
         self.set_num_panes(len(locations))
         locations = [os.path.abspath(l or ".") for l in locations]
+        self.current_path = None
         self.model.clear()
         for pane, loc in enumerate(locations):
             self.fileentry[pane].set_filename(loc)
@@ -475,6 +520,12 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
             path = todo.pop(0)
             it = self.model.get_iter( path )
             roots = self.model.value_paths( it )
+
+            # Buggy ordering when deleting rows means that we sometimes try to
+            # recursively update files; this fix seems the least invasive.
+            if not any(os.path.isdir(root) for root in roots):
+                continue
+
             yield _("[%s] Scanning %s") % (self.label_text, roots[0][prefixlen:])
             differences = False
             encoding_errors = []
@@ -491,7 +542,7 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
 
                 try:
                     entries = os.listdir(root)
-                except OSError, err:
+                except OSError as err:
                     self.model.add_error(it, err.strerror, pane)
                     differences = True
                     continue
@@ -504,7 +555,7 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
                 for e in entries:
                     try:
                         e = e.decode('utf8')
-                    except UnicodeDecodeError, err:
+                    except UnicodeDecodeError:
                         approximate_name = e.decode('utf8', 'replace')
                         encoding_errors.append((pane, approximate_name))
                         continue
@@ -512,7 +563,7 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
                     try:
                         s = os.lstat(os.path.join(root, e))
                     # Covers certain unreadable symlink cases; see bgo#585895
-                    except OSError, err:
+                    except OSError as err:
                         error_string = e + err.strerror
                         self.model.add_error(it, error_string, pane)
                         continue
@@ -530,7 +581,7 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
                                 files.add(pane, e)
                             elif stat.S_ISDIR(s.st_mode):
                                 dirs.add(pane, e)
-                        except OSError, err:
+                        except OSError as err:
                             if err.errno == errno.ENOENT:
                                 error_string = e + ": Dangling symlink"
                             else:
@@ -677,7 +728,7 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
                                 continue
                         misc.copytree(src, dst)
                         self.recursively_update( path )
-                except (OSError,IOError), e:
+                except (OSError, IOError) as e:
                     misc.run_dialog(_("Error copying '%s' to '%s'\n\n%s.") % (src, dst,e), self)
 
     def delete_selected(self):
@@ -700,11 +751,60 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
                                 parent = self,
                                 buttonstype=gtk.BUTTONS_OK_CANCEL) == gtk.RESPONSE_OK:
                             shutil.rmtree(name)
-                            self.recursively_update( path )
-                        self.file_deleted( path, pane)
-                except OSError, e:
+                            self.recursively_update(path)
+                            self.file_deleted(path, pane)
+                except OSError as e:
                     misc.run_dialog(_("Error removing %s\n\n%s.") % (name,e), parent = self)
 
+    def on_treemodel_row_deleted(self, model, path):
+
+        # TODO: Move this and path tools to new tree helper module
+        def refocus_deleted_path(model, path):
+            # Since the passed path has been deleted, either the path is now a
+            # valid successor, or there are no successors. If valid, return it.
+            # If not, and the path has a predecessor sibling (immediate or
+            # otherwise), then return that. If there are no siblings, traverse
+            # parents until we get a valid path, and return that.
+
+            def tree_path_prev(path):
+                if not path or path[-1] == 0:
+                    return None
+                return path[:-1] + (path[-1] - 1,)
+
+            def tree_path_up(path):
+                if not path:
+                    return None
+                return path[:-1]
+
+            def valid_path(model, path):
+                try:
+                    model.get_iter(path)
+                    return True
+                except ValueError:
+                    return False
+
+            if valid_path(model, path):
+                return path
+
+            new_path = tree_path_prev(path)
+            while new_path:
+                if valid_path(model, new_path):
+                    return new_path
+                new_path = tree_path_prev(new_path)
+
+            new_path = tree_path_up(path)
+            while new_path:
+                if valid_path(model, new_path):
+                    return new_path
+                new_path = tree_path_up(new_path)
+
+            return None
+
+        if self.current_path == path:
+            self.current_path = refocus_deleted_path(model, path)
+            if self.current_path and self.treeview_focussed:
+                self.treeview_focussed.set_cursor(self.current_path)
+
     def on_treeview_cursor_changed(self, *args):
         pane = self._get_focused_pane()
         if pane is None:
@@ -714,15 +814,17 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
         if not cursor_path:
             self.emit("next-diff-changed", False, False)
         else:
-            if self.current_path:
+            try:
                 old_cursor = self.model.get_iter(self.current_path)
+            except (ValueError, TypeError):
+                # An invalid path gives ValueError; None gives a TypeError
+                skip = False
+            else:
                 state = self.model.get_state(old_cursor, 0)
                 # We can skip recalculation if the new cursor is between the
                 # previous/next bounds, and we weren't on a changed row
                 skip = state in (tree.STATE_NORMAL, tree.STATE_EMPTY) and \
                        self.prev_path < cursor_path < self.next_path
-            else:
-                skip = False
 
             if not skip:
                 prev, next = self.model._find_next_prev_diff(cursor_path)
@@ -753,9 +855,13 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
                 stat = os.stat(fname)
             # TypeError for if fname is None
             except (OSError, TypeError):
-                self.emit("status-changed", "")
+                self.status_info_labels[0].set_text("")
+                self.status_info_labels[1].set_markup("")
             else:
-                self.emit("status-changed", "%s : %s" % (rwx(stat.st_mode), nice(time.time() - stat.st_mtime) ) )
+                mode_text = "<tt>%s</tt>" % rwx(stat.st_mode)
+                last_changed_text = str(nice(time.time() - stat.st_mtime))
+                self.status_info_labels[0].set_text(last_changed_text)
+                self.status_info_labels[1].set_markup(mode_text)
 
     def on_treeview_key_press_event(self, view, event):
         pane = self.treeview.index(view)
@@ -840,9 +946,12 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
 
     def open_external(self):
         pane = self._get_focused_pane()
-        if pane is not None:
-            m = self.model
-            files = [ m.value_path( m.get_iter(p), pane ) for p in self._get_selected_paths(pane) ]
+        if pane is None:
+            return
+        path = lambda p: self.model.value_path(self.model.get_iter(p), pane)
+        files = [path(p) for p in self._get_selected_paths(pane)]
+        files = [f for f in files if f]
+        if files:
             self._open_files(files)
 
     def on_button_ignore_case_toggled(self, button):
@@ -943,26 +1052,27 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
                 isdir = os.path.isdir( files[j] )
                 # TODO: Differentiate the DodgySame case
                 if all_same == Same or all_same == DodgySame:
-                    self.model.set_state(it, j,  tree.STATE_NORMAL, isdir)
+                    self.model.set_path_state(it, j, tree.STATE_NORMAL, isdir)
                     different = 0
                 elif all_same == SameFiltered:
-                    self.model.set_state(it, j,  tree.STATE_NOCHANGE, isdir)
+                    self.model.set_path_state(it, j, tree.STATE_NOCHANGE, isdir)
                     different = 0
                 # TODO: Differentiate the SameFiltered and DodgySame cases
                 elif all_present_same in (Same, SameFiltered, DodgySame):
-                    self.model.set_state(it, j,  tree.STATE_NEW, isdir)
+                    self.model.set_path_state(it, j, tree.STATE_NEW, isdir)
                 elif all_same == FileError or all_present_same == FileError:
-                    self.model.set_state(it, j,  tree.STATE_ERROR, isdir)
+                    self.model.set_path_state(it, j, tree.STATE_ERROR, isdir)
                 # Different and DodgyDifferent
                 else:
-                    self.model.set_state(it, j,  tree.STATE_MODIFIED, isdir)
+                    self.model.set_path_state(it, j, tree.STATE_MODIFIED, isdir)
                 self.model.set_value(it,
                     self.model.column_index(COL_EMBLEM, j),
                     j == newest_index and "emblem-meld-newer-file" or None)
                 one_isdir[j] = isdir
         for j in range(self.model.ntree):
             if not mod_times[j]:
-                self.model.set_state(it, j, tree.STATE_MISSING, True in one_isdir)
+                self.model.set_path_state(it, j, tree.STATE_NONEXIST,
+                                          True in one_isdir)
         return different
 
     def popup_in_pane(self, pane, event):
@@ -991,6 +1101,8 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
         if event.button == 3:
             treeview.grab_focus()
             path = treeview.get_path_at_pos(int(event.x), int(event.y))
+            if path is None:
+                return False
             selection = treeview.get_selection()
             model, rows = selection.get_selected_rows()
 
@@ -1031,19 +1143,12 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
             self.model = DirDiffTreeStore(n)
             for i in range(n):
                 self.treeview[i].set_model(self.model)
-
-            colour_map = {
-                "conflict": (1.0, 0.75294117647058822, 0.79607843137254897),
-                "error": (0.9882352941176, 0.9137254901960, 0.30980392156862),
-                "insert": (0.75686274509803919, 1.0, 0.75686274509803919),
-                "replace": (0.8666666666666667, 0.93333333333333335, 1.0),
-                "delete": (1.0, 1.0, 1.0),
-            }
+            self.model.connect("row-deleted", self.on_treemodel_row_deleted)
 
             for (w, i) in zip(self.diffmap, (0, n - 1)):
                 scroll = self.scrolledwindow[i].get_vscrollbar()
                 idx = 1 if i else 0
-                w.setup(scroll, self.get_state_traversal(idx), colour_map)
+                w.setup(scroll, self.get_state_traversal(idx), [self.fill_colors, self.line_colors])
 
             toshow =  self.scrolledwindow[:n] + self.fileentry[:n]
             toshow += self.linkmap[:n-1] + self.diffmap[:n]
@@ -1131,3 +1236,9 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
 
     def on_reload_activate(self, *extra):
         self.on_fileentry_activate(None)
+
+    def on_delete_event(self, appquit=0):
+        for h in self.app_handlers:
+            app.disconnect(h)
+
+        return gtk.RESPONSE_OK
