@@ -27,7 +27,7 @@ import gtk
 import misc
 import preferences
 
-version = "1.5.4"
+version = "1.7.0"
 
 
 class FilterEntry(object):
@@ -88,7 +88,7 @@ class FilterEntry(object):
         elif filter_type == FilterEntry.SHELL:
             compiled = FilterEntry._compile_shell_pattern(filter_string)
         else:
-            raise ValueError, "Unknown filter type"
+            raise ValueError("Unknown filter type")
         return compiled
 
     def __copy__(self):
@@ -147,7 +147,7 @@ class MeldApp(gobject.GObject):
                 diff_files_args.append(arg)
                 del parser.rargs[0]
 
-        if len(diff_files_args) not in (1, 2, 3, 4):
+        if len(diff_files_args) not in (1, 2, 3):
             raise optparse.OptionValueError(
                 _("wrong number of arguments supplied to --diff"))
         parser.values.diff.append(diff_files_args)
@@ -167,6 +167,8 @@ class MeldApp(gobject.GObject):
             version="%prog " + version)
         parser.add_option("-L", "--label", action="append", default=[],
             help=_("Set label to use instead of file name"))
+        parser.add_option("-n", "--newtab", action="store_true", default=False,
+            help=_("Open a new tab in an already running instance"))
         parser.add_option("-a", "--auto-compare", action="store_true", default=False,
             help=_("Automatically compare all differing files on startup"))
         parser.add_option("-u", "--unified", action="store_true",
@@ -174,27 +176,42 @@ class MeldApp(gobject.GObject):
         parser.add_option("-o", "--output", action="store", type="string",
             dest="outfile", default=None,
             help=_("Set the target file for saving a merge result"))
+        parser.add_option("--auto-merge", None, action="store_true", default=False,
+            help=_("Automatically merge files"))
         parser.add_option("", "--diff", action="callback", callback=self.diff_files_callback,
                           dest="diff", default=[],
                           help=_("Creates a diff tab for up to 3 supplied files or directories."))
         options, args = parser.parse_args(rawargs)
-        if len(args) > 4:
-            parser.error(_("too many arguments (wanted 0-4, got %d)") % len(args))
-        elif len(args) == 4 and any([os.path.isdir(f) for f in args]):
-            parser.error(_("can't compare more than three directories"))
+        if len(args) > 3:
+            parser.error(_("too many arguments (wanted 0-3, got %d)") % len(args))
+        elif options.auto_merge and len(args) < 3:
+            parser.error(_("can't auto-merge less than 3 files"))
+        elif options.auto_merge and any([os.path.isdir(f) for f in args]):
+            parser.error(_("can't auto-merge directories"))
+
+        new_window = True
+        open_paths = self.window.open_paths
+        if options.newtab:
+            if not dbus_app:
+                print _("D-Bus error; comparisons will open in a new window.")
+            else:
+                # Note that we deliberately discard auto-compare and -merge
+                # options here; these are not supported via dbus yet.
+                open_paths = lambda f, *x: dbus_app.OpenPaths(f, 0)
+                new_window = False
 
         for files in options.diff:
-            if len(files) == 4 and any([os.path.isdir(f) for f in files]):
-                parser.error(_("can't compare more than three directories"))
-            self.window.open_paths(files)
+            open_paths(files)
 
-        tab = self.window.open_paths(args, options.auto_compare)
+        tab = open_paths(args, options.auto_compare, options.auto_merge)
         if options.label and tab:
             tab.set_labels(options.label)
 
         if options.outfile and tab and isinstance(tab, filediff.FileDiff):
             tab.set_merge_output_file(options.outfile)
 
+        return new_window
+
 
 app = MeldApp()
 
