@@ -16,14 +16,16 @@
 ### Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
 
+import os
 import subprocess
 import sys
 
 import gobject
-import task
-import undo
+import gio
 import gtk
-import os
+
+import task
+
 from gettext import gettext as _
 
 
@@ -48,13 +50,16 @@ class MeldDoc(gobject.GObject):
 
     def __init__(self, prefs):
         gobject.GObject.__init__(self)
-        self.undosequence = undo.UndoSequence()
         self.scheduler = task.FifoScheduler()
         self.prefs = prefs
         self.prefs.notify_add(self.on_preference_changed)
         self.num_panes = 0
         self.label_text = _("untitled")
         self.tooltip_text = _("untitled")
+        self.status_info_labels = []
+
+    def get_info_widgets(self):
+        return self.status_info_labels
 
     def save(self):
         pass
@@ -67,38 +72,45 @@ class MeldDoc(gobject.GObject):
             self.scheduler.remove_task(self.scheduler.get_current_task())
 
     def _open_files(self, selected):
-        files = [f for f in selected if os.path.isfile(f)]
-        dirs = [d for d in selected if os.path.isdir(d)]
+        query_attrs = ",".join((gio.FILE_ATTRIBUTE_STANDARD_TYPE,
+                                gio.FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE))
 
-        def os_open(paths):
-            for path in paths:
-                if sys.platform == "win32":
-                    subprocess.Popen(["start", path], shell=True)
-                elif sys.platform == "darwin":
-                    subprocess.Popen(["open", path])
-                else:
-                    subprocess.Popen(["xdg-open", path])
-
-        if len(files):
-            cmd = self.prefs.get_editor_command(files)
-            if cmd:
-                os.spawnvp(os.P_NOWAIT, cmd[0], cmd)
+        def os_open(path):
+            if not path:
+                return
+            if sys.platform == "win32":
+                subprocess.Popen(["start", path], shell=True)
+            elif sys.platform == "darwin":
+                subprocess.Popen(["open", path])
             else:
-                os_open(files)
+                subprocess.Popen(["xdg-open", path])
 
-        os_open(dirs)
+        def open_cb(source, result, *data):
+            info = source.query_info_finish(result)
+            file_type = info.get_file_type()
+            if file_type == gio.FILE_TYPE_DIRECTORY:
+                os_open(source.get_path())
+            elif file_type == gio.FILE_TYPE_REGULAR:
+                content_type = info.get_content_type()
+                path = source.get_path()
+                if gio.content_type_is_a(content_type, "text/plain"):
+                    editor = self.prefs.get_editor_command([path])
+                    if editor:
+                        subprocess.Popen(editor)
+                    else:
+                        os_open(path)
+                else:
+                    os_open(path)
+            else:
+                # TODO: Add some kind of 'failed to open' notification
+                pass
+
+        for f in [gio.File(s) for s in selected]:
+            f.query_info_async(query_attrs, open_cb)
 
     def open_external(self):
         pass
 
-    def on_undo_activate(self):
-        if self.undosequence.can_undo():
-            self.undosequence.undo()
-
-    def on_redo_activate(self):
-        if self.undosequence.can_redo():
-            self.undosequence.redo()
-
     def on_refresh_activate(self, *extra):
         self.on_reload_activate(self, *extra)
 
