@@ -135,6 +135,9 @@ def make_tool_button_widget(label):
     hbox.show_all()
     return hbox
 
+def gdk_to_cairo_color(color):
+    return (color.red / 65535., color.green / 65535., color.blue / 65535.)
+
 def all_equal(alist):
     """Return true if all members of the list are equal to the first.
 
@@ -254,6 +257,8 @@ def copy2(src, dst):
         dst = os.path.join(dst, os.path.basename(src))
 
     if os.path.islink(src) and os.path.isfile(src):
+        if os.path.lexists(dst):
+            os.unlink(dst)
         os.symlink(os.readlink(src), dst)
     elif os.path.isfile(src):
         shutil.copyfile(src, dst)
@@ -262,7 +267,7 @@ def copy2(src, dst):
 
     try:
         shutil.copystat(src, dst)
-    except OSError, e:
+    except OSError as e:
         if e.errno != errno.EPERM:
             raise
 
@@ -277,7 +282,7 @@ def copytree(src, dst):
 
     try:
         os.mkdir(dst)
-    except OSError, e:
+    except OSError as e:
         if e.errno != errno.EEXIST:
             raise
     names = os.listdir(src)
@@ -293,7 +298,7 @@ def copytree(src, dst):
 
     try:
         shutil.copystat(src, dst)
-    except OSError, e:
+    except OSError as e:
         if e.errno != errno.EPERM:
             raise
 
