@@ -53,7 +53,7 @@ class Vc(_vc.CachedVc):
     def update_command(self):
         return [self.CMD, "update"]
 
-    def add_command(self, binary=0):
+    def add_command(self):
         return [self.CMD, "add"]
 
     def remove_command(self, force=0):
@@ -84,7 +84,7 @@ class Vc(_vc.CachedVc):
                 entries = _vc.popen([self.CMD, "ls", "-l"],
                                     cwd=self.root).read().split("\n")[:-1]
                 break
-            except OSError, e:
+            except OSError as e:
                 if e.errno != errno.EAGAIN:
                     raise
 
@@ -139,7 +139,7 @@ class Vc(_vc.CachedVc):
                                       'unchanged']:
                         rev = entries[1].strip()
                     break
-                except OSError, e:
+                except OSError as e:
                     if e.errno != errno.EAGAIN:
                         raise
 
