@@ -17,10 +17,12 @@
 
 import tempfile
 import shutil
-import gtk
 import os
 from gettext import gettext as _
 
+import gtk
+import pango
+
 import tree
 import misc
 from ui import gnomeglade
@@ -80,10 +82,7 @@ COL_LOCATION, COL_STATUS, COL_REVISION, COL_TAG, COL_OPTIONS, COL_END = range(tr
 
 class VcTreeStore(tree.DiffTreeStore):
     def __init__(self):
-        ntree = 1
-        types = [str] * COL_END * ntree
-        tree.DiffTreeStore.__init__(self, ntree, types)
-        self.textstyle[tree.STATE_MISSING] = '<span foreground="#000088" strikethrough="true" weight="bold">%s</span>'
+        tree.DiffTreeStore.__init__(self, 1, [str] * 5)
 
 ################################################################################
 # filters
@@ -105,7 +104,6 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
                          "VcCommit": ("commit_command", ("",)),
                          "VcUpdate": ("update_command", ()),
                          "VcAdd": ("add_command", ()),
-                         "VcAddBinary": ("add_command", ()),
                          "VcResolved": ("resolved_command", ()),
                          "VcRemove": ("remove_command", ()),
                          "VcRevert": ("revert_command", ()),
@@ -128,7 +126,6 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
             ("VcCommit",        "vc-commit-24",             _("Co_mmit"),       None, _("Commit"), self.on_button_commit_clicked),
             ("VcUpdate",        "vc-update-24",             _("_Update"),       None, _("Update"), self.on_button_update_clicked),
             ("VcAdd",           "vc-add-24",                _("_Add"),          None, _("Add to VC"), self.on_button_add_clicked),
-            ("VcAddBinary",     None,                       _("Add _Binary"),   None, _("Add binary to VC"), self.on_button_add_binary_clicked),
             ("VcRemove",        "vc-remove-24",             _("_Remove"),       None, _("Remove from VC"), self.on_button_remove_clicked),
             ("VcResolved",      "vc-resolve-24",            _("_Resolved"),     None, _("Mark as resolved for VC"), self.on_button_resolved_clicked),
             ("VcRevert",        gtk.STOCK_REVERT_TO_SAVED,  None,               None, _("Revert to original"), self.on_button_revert_clicked),
@@ -158,6 +155,7 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
             button.props.icon_name = button.props.stock_id
         self.tempdirs = []
         self.model = VcTreeStore()
+        self.widget.connect("style-set", self.model.on_style_set)
         self.treeview.set_model(self.model)
         self.treeview.get_selection().set_mode(gtk.SELECTION_MULTIPLE)
         self.treeview.set_headers_visible(1)
@@ -172,7 +170,13 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
         column.set_attributes(renicon,
                               icon_name=col_index(tree.COL_ICON, 0),
                               icon_tint=col_index(tree.COL_TINT, 0))
-        column.set_attributes(rentext, markup=col_index(tree.COL_TEXT, 0))
+        column.set_attributes(rentext,
+                    text=col_index(tree.COL_TEXT, 0),
+                    foreground_gdk=col_index(tree.COL_FG, 0),
+                    style=col_index(tree.COL_STYLE, 0),
+                    weight=col_index(tree.COL_WEIGHT, 0),
+                    strikethrough=col_index(tree.COL_STRIKE, 0))
+
         self.treeview.append_column(column)
 
         def addCol(name, num):
@@ -210,7 +214,6 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
         self.consolestream = ConsoleStream(self.consoleview)
         self.location = None
         self.treeview_column_location.set_visible(self.actiongroup.get_action("VcFlatten").get_active())
-        self.fileentry.show() #TODO: remove once bug 97503 is fixed
         if not self.prefs.vc_console_visible:
             self.on_console_view_toggle(self.console_hide_box)
         self.vc = None
@@ -288,7 +291,7 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
     def on_vc_change(self, cb):
         if not cb.lock:
             self.vc = cb.get_model()[cb.get_active_iter()][1]
-            self._set_location(self.vc.root)
+            self._set_location(self.vc.location)
             self.update_actions_sensitivity()
 
     def set_location(self, location):
@@ -296,13 +299,14 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
 
     def _set_location(self, location):
         self.location = location
+        self.current_path = None
         self.model.clear()
         self.fileentry.set_filename(location)
         self.fileentry.prepend_history(location)
         it = self.model.add_entries( None, [location] )
         self.treeview.grab_focus()
         self.treeview.get_selection().select_iter(it)
-        self.model.set_state(it, 0, tree.STATE_NORMAL, isdir=1)
+        self.model.set_path_state(it, 0, tree.STATE_NORMAL, isdir=1)
         self.recompute_label()
         self.scheduler.remove_all_tasks()
 
@@ -405,6 +409,12 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
                 diff = difffunc()
                 yield 1
             prefix, patch = diff[0], diff[1]
+
+            try:
+                patch = self.vc.clean_patch(patch)
+            except AttributeError:
+                pass
+
             yield _("[%s] Applying patch") % self.label_text
             if patch:
                 applied = self.show_patch(prefix, patch, silent=silent_error)
@@ -428,6 +438,8 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
     def on_button_press_event(self, treeview, event):
         if event.button == 3:
             path = treeview.get_path_at_pos(int(event.x), int(event.y))
+            if path is None:
+                return False
             selection = treeview.get_selection()
             model, rows = selection.get_selected_rows()
 
@@ -490,7 +502,7 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
                 r = readfunc()
                 self.consolestream.write(r)
                 yield 1
-        except IOError, e:
+        except IOError as e:
             misc.run_dialog("Error running command.\n'%s'\n\nThe error was:\n%s" % ( misc.shelljoin(command), e),
                 parent=self, messagetype=gtk.MESSAGE_ERROR)
         if refresh:
@@ -516,9 +528,8 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
         dialog.run()
 
     def on_button_add_clicked(self, obj):
-        self._command_on_selected(self.vc.add_command() )
-    def on_button_add_binary_clicked(self, obj):
-        self._command_on_selected(self.vc.add_command(binary=1))
+        self._command_on_selected(self.vc.add_command())
+
     def on_button_remove_clicked(self, obj):
         self._command_on_selected(self.vc.remove_command())
     def on_button_resolved_clicked(self, obj):
@@ -536,7 +547,7 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
                             parent = self,
                             buttonstype=gtk.BUTTONS_OK_CANCEL) == gtk.RESPONSE_OK:
                         shutil.rmtree(name)
-            except OSError, e:
+            except OSError as e:
                 misc.run_dialog(_("Error removing %s\n\n%s.") % (name,e), parent = self)
         workdir = _commonprefix(files)
         self.refresh_partial(workdir)
@@ -550,6 +561,19 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
         self._open_files(self._get_selected_files())
 
     def show_patch(self, prefix, patch, silent=False):
+        if vc._vc.call(["which", "patch"]):
+            primary = _("Patch tool not found")
+            secondary = _("Meld needs the <i>patch</i> tool to be installed "
+                          "to perform comparisons in %s repositories. Please "
+                          "install <i>patch</i> and try again.") % self.vc.NAME
+
+            msgarea = self.msgarea_mgr.new_from_text_and_icon(
+                        gtk.STOCK_DIALOG_ERROR, primary, secondary)
+            msgarea.add_button(_("Hi_de"), gtk.RESPONSE_CLOSE)
+            msgarea.connect("response", lambda *args: self.msgarea_mgr.clear())
+            msgarea.show_all()
+            return False
+
         tmpdir = tempfile.mkdtemp("-meld")
         self.tempdirs.append(tmpdir)
 
@@ -579,35 +603,28 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
                 self.emit("create-diff", d)
             return True
         elif not silent:
-            import meldapp
-            msg = _("""
-                    Invoking 'patch' failed.
-                    
-                    Maybe you don't have 'GNU patch' installed,
-                    or you use an untested version of %s.
-                    
-                    Please send email bug report to:
-                    meld-list@gnome.org
-                    
-                    Containing the following information:
-                    
-                    - meld version: '%s'
-                    - source control software type: '%s'
-                    - source control software version: 'X.Y.Z'
-                    - the output of '%s somefile.txt'
-                    - patch command: '%s'
-                    (no need to actually run it, just provide
-                    the command line) 
-                    
-                    Replace 'X.Y.Z' by the actual version for the
-                    source control software you use.
-                    """) % (self.vc.NAME,
-                            meldapp.version,
-                            self.vc.NAME,
-                            " ".join(self.vc.diff_command()),
-                            " ".join(patchcmd))
-            msg = '\n'.join([line.strip() for line in msg.split('\n')])
-            misc.run_dialog(msg, parent=self)
+            primary = _("Error fetching original comparison file")
+            secondary = _("Meld couldn't obtain the original version of your "
+                          "comparison file. If you are using the most recent "
+                          "version of Meld, please report a bug, including as "
+                          "many details as possible.")
+
+            msgarea = self.msgarea_mgr.new_from_text_and_icon(
+                        gtk.STOCK_DIALOG_ERROR, primary, secondary)
+            msgarea.add_button(_("Hi_de"), gtk.RESPONSE_CLOSE)
+            msgarea.add_button(_("Report a bug"), gtk.RESPONSE_OK)
+
+            def patch_error_cb(msgarea, response):
+                if response == gtk.RESPONSE_OK:
+                    bug_url = "https://bugzilla.gnome.org/enter_bug.cgi?" + \
+                              "product=meld"
+                    misc.open_uri(bug_url)
+                else:
+                    self.msgarea_mgr.clear()
+
+            msgarea.connect("response", patch_error_cb)
+            msgarea.show_all()
+
         return False
 
     def refresh(self):
@@ -619,7 +636,7 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
             if it:
                 newiter = self.model.insert_after( None, it)
                 self.model.set_value(newiter, self.model.column_index( tree.COL_PATH, 0), where)
-                self.model.set_state(newiter, 0, tree.STATE_NORMAL, isdir=1)
+                self.model.set_path_state(newiter, 0, tree.STATE_NORMAL, True)
                 self.model.remove(it)
                 self.scheduler.add_task( self._search_recursively_iter(newiter).next )
         else: # XXX fixme
@@ -627,7 +644,7 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
 
     def _update_item_state(self, it, vcentry, location):
         e = vcentry
-        self.model.set_state( it, 0, e.state, e.isdir )
+        self.model.set_path_state(it, 0, e.state, e.isdir)
         def setcol(col, val):
             self.model.set_value(it, self.model.column_index(col, 0), val)
         setcol(COL_LOCATION, location)
@@ -696,15 +713,17 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
         if not cursor_path:
             self.emit("next-diff-changed", False, False)
         else:
-            if self.current_path:
+            try:
                 old_cursor = self.model.get_iter(self.current_path)
+            except (ValueError, TypeError):
+                # An invalid path gives ValueError; None gives a TypeError
+                skip = False
+            else:
                 state = self.model.get_state(old_cursor, 0)
                 # We can skip recalculation if the new cursor is between the
                 # previous/next bounds, and we weren't on a changed row
                 skip = state in (tree.STATE_NORMAL, tree.STATE_EMPTY) and \
                        self.prev_path < cursor_path < self.next_path
-            else:
-                skip = False
 
             if not skip:
                 prev, next = self.model._find_next_prev_diff(cursor_path)
